/**
 *
 * WapiDAO.java
 *
 * Copyright 2012 Infoblox Inc. All rights reserved.
 *             
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *     1. Redistributions of source code must retain the above copyright notice,
 *     this list of conditions and the following disclaimer.
 *     2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY INFOBLOX AND CONTRIBUTORS ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL INFOBLOX OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * The views and conclusions contained in the software and documentation are
 * those of the authors and should not be interpreted as representing official
 * policies, either expressed or implied, of Infoblox.
 *
 */
package com.infoblox.wapi.dao;

import org.apache.log4j.Logger;

import com.infoblox.wapi.common.WapiConfig.ContentType;
import com.infoblox.wapi.common.WapiConfig.ReturnType;
import com.infoblox.wapi.common.WapiPort;
import com.infoblox.wapi.common.WapiPort.RequestMethod;
import com.infoblox.wapi.common.WapiResponse;
import com.infoblox.wapi.converter.AbstractConverter;

/**
 * Default implementation of IWapiDao interface.
 * 
 * @param <I>
 *            Input class: JSONObject, Document or String..
 * @param <O>
 *            Output class: JSONArray or Document.
 */
public class WapiDAO<I, O> implements IWapiDAO<I, O> {

    /** Log4j logger */
    private static Logger logger = Logger.getLogger(WapiDAO.class);

    /** Current object type */
    private String objectType = "network"; //$NON-NLS-1$

    /** WapiPort for this DAO */
    protected WapiPort port;

    /** Input class: JSONObject, Document or String. */
    protected Class<I> input;

    /** Output class: JSONArray or Document. */
    protected Class<O> output;

    /**
     * @param newPort
     * @param newObjectType
     * @param newInput
     * @param newOutput
     */
    public WapiDAO(WapiPort newPort, String newObjectType, Class<I> newInput,
            Class<O> newOutput) {
        this.objectType = newObjectType;
        this.port = newPort;
        this.input = newInput;
        this.output = newOutput;
        port.getConfig().setContentType(ContentType.getInstanceByClass(input));
        port.getConfig().setReturnType(ReturnType.getInstanceByClass(output));
    }

    /**
     * @param response
     * @return Handling responses with errors.
     */
    public String getResponse(WapiResponse response) {
        if (200 <= response.getCode() && response.getCode() < 300) {
            return response.getResponseText();
        }
        logger.error("Got response with error:\n" + response.getResponseText()); //$NON-NLS-1$
        throw new RuntimeException(
                "Got response with error:\n" + response.getResponseText()); //$NON-NLS-1$
    }

    /**
     * @param string
     * @return Converted object.
     */
    @SuppressWarnings("unchecked")
    private O convertToObject(String string) {
        AbstractConverter<O> converter = (AbstractConverter<O>) port
                .getConfig().getReturnType().getConverter();
        return converter.stringToObject(string);
    }

    /**
     * @param object
     * @return String
     */
    private String convertToString(I object) {
        @SuppressWarnings("unchecked")
        AbstractConverter<I> converter = (AbstractConverter<I>) port
                .getConfig().getContentType().getConverter();
        return converter.objectToString(object);
    }

    /**
     * @param newRef
     * @return Object which was found by ref.
     * @see com.infoblox.wapi.dao.IWapiDAO#read(java.lang.String)
     */
    @Override
    public O read(String newRef) {
        String ref = newRef;
        if (null == ref) {
            ref = objectType;
        }
        return convertToObject(getResponse(port.sendRequest(ref,
                RequestMethod.GET)));
    }

    /**
     * Read object by ref in URL.
     * 
     * @return Object which was found by ref in URL which was set in the config.
     * @see com.infoblox.wapi.dao.IWapiDAO#read(java.lang.String)
     */
    @Override
    public O read() {
        return this.read(null);
    }

    /**
     * @return All objects with specified object type.
     * @see com.infoblox.wapi.dao.IWapiDAO#readAll()
     */
    @Override
    public O readAll() {
        return convertToObject(getResponse(port.sendRequest(objectType,
                RequestMethod.GET)));
    }

    /**
     * @param object
     * @return String with ref of created object.
     * @see com.infoblox.wapi.dao.IWapiDAO#create(Object)
     */
    @Override
    public String create(I object) {
        String response = getResponse(port.sendRequest(objectType,
                RequestMethod.POST, convertToString(object)));

        return response.replaceAll("^\"|\"$", ""); //$NON-NLS-1$ //$NON-NLS-2$
    }

    /**
     * @param newRef
     * @param object
     * @return String with ref of updated object.
     * @see com.infoblox.wapi.dao.IWapiDAO#update(java.lang.String, Object)
     */
    @Override
    public String update(String newRef, I object) {
        String ref = newRef;
        if (null == ref) {
            ref = objectType;
        }
        String response = getResponse(port.sendRequest(ref, RequestMethod.PUT,
                convertToString(object)));
        return response.replaceAll("^\"|\"$", ""); //$NON-NLS-1$ //$NON-NLS-2$

    }

    /**
     * Update object by ref in URL.
     * 
     * @param object
     * @return String with ref of updated object.
     */
    @Override
    public String update(I object) {
        return this.update(null, object);
    }

    /**
     * @param newRef
     * @return String with ref of deleted object.
     * @see com.infoblox.wapi.dao.IWapiDAO#delete(java.lang.String)
     */
    @Override
    public String delete(String newRef) {
        String ref = newRef;
        if (null == ref) {
            ref = objectType;
        }
        String response = getResponse(port.sendRequest(ref,
                RequestMethod.DELETE));
        return response.replaceAll("^\"|\"$", ""); //$NON-NLS-1$ //$NON-NLS-2$
    }

    /**
     * Delete object by ref in URL which was set in the config.
     * 
     * @return String with ref of deleted object.
     */
    @Override
    public String delete() {
        return this.delete(null);
    }

    /**
     * @param object
     * @return Objects that satisfy the search query.
     * @see com.infoblox.wapi.dao.IWapiDAO#search(Object)
     */
    @Override
    public O search(I object) {
        String response = getResponse(port.sendRequest(objectType,
                "_method=GET", //$NON-NLS-1$
                RequestMethod.POST, convertToString(object)));
        return convertToObject(response.replaceAll("^\"|\"$", "")); //$NON-NLS-1$ //$NON-NLS-2$
    }

    /**
     * @return AbstractConverter for input objects.
     */
    @Override
    @SuppressWarnings("unchecked")
    public AbstractConverter<I> getInputConverter() {
        return (AbstractConverter<I>) port.getConfig().getContentType()
                .getConverter();
    }

    /**
     * @return AbstractConverter for output objects.
     */
    @Override
    @SuppressWarnings("unchecked")
    public AbstractConverter<O> getOutputConverter() {
        return (AbstractConverter<O>) port.getConfig().getReturnType()
                .getConverter();
    }

}